from typing import List, Callable, Tuple, Any, Dict, Union
from numbers import Number
from collections import defaultdict
import numpy as np

from ._markov_chain import MarkovChain, BaseMarkovChain
from .samplers import VanillaSampler, Sampler
from .parameterization import Parameterization
from ._target import Target
from ._state import State
from ._utils import _preprocess_func, _LogLikeRatioFromFunc


class BaseBayesianInversion:
    r"""
    A low-level class for Bayesian sampling based on Markov chain Monte Carlo (McMC).

    This class provides the basic structure for setting up and running McMC sampling,
    given user-provided definition of perturbations, likelihood functions and the
    initialization of walkers. At each iteration of the inference process,
    the current model :math:`\bf m` is perturbed to produce :math:`\bf m'`, and
    the new model is accepted with probability

    .. math::

        \alpha({\bf m' \mid m}) = \mbox{min} \Bigg[1,
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}}
            \underbrace{\frac{p\left({\bf d}_{obs} \mid {\bf m'}\right)}{p\left({\bf d}_{obs} \mid {\bf m}\right)}}_{\text{Likelihood ratio}}
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}
        \Bigg],

    where :math:`p({\bf d}_{obs})` denotes the observed data and
    :math:`\mathbf{J}` the Jacobian of the transformation.

    Parameters
    ----------
    walkers_starting_models: List[Any]
        a list of starting models for each chain. The models can be of any type so long
        as they are consistent with what is accepted as arguments in the perturbation
        functions and probability functions. The length of this list must be equal to
        the number of chains, i.e. ``n_chains``
    perturbation_funcs: List[Callable[[Any], Tuple[Any, Number]]]
        a list of perturbation functions. Each perturbation function should take
        in a model :math:`\mathbf{m}` (any type is allowed, as long as it is
        consistent with ``walkers_starting_models`` and the below functions) and
        perturb it to produce the new model :math:`\bf m'`. Each perturbation function
        should return :math:`\bf m'` along with :math:`\log(
        \frac{p({\bf m'})}{p({\bf m})}
        \frac{q\left({\bf m}
        \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
        \lvert \mathbf{J} \rvert)`, which is used in the calculation of
        the acceptance probability.
    log_likelihood_func: Callable[[Any], Number], optional
        the log likelihood function :math:`\log(p(\mathbf{d}_{obs} \mid \mathbf{m}))`.
        It takes in a model :math:`\mathbf{m}` (any type is allowed, as long as it is
        consistent with the other arguments of this class) and returns the log
        of the likelihood function. This function is only used when ``log_like_ratio_func``
        is None. Default is None
    log_like_ratio_func: Callable[[Any, Any], Number], optional
        the log likelihood ratio function :math:`\log(\frac{p(\mathbf{d}_{obs} \mid \mathbf{m'})}
        {p(\mathbf{d}_{obs} \mid \mathbf{m})})`. It takes the current and proposed models,
        :math:`\mathbf{m}` and :math:`\mathbf{m'}`, whose type should be consistent
        with the other arguments of this class, and returns a scalar corresponding to
        the log likelihood ratio. This is utilised in the calculation of the
        acceptance probability. If None, ``log_likelihood_func`` gets used instead. Default is None
    n_chains: int, optional
        the number of chains in the McMC sampling. Default is 10
    n_cpus: int, optional
        the number of CPUs available. This is usually set to be equal to the number of
        chains if there are enough CPUs. Default is 10
    """

    def __init__(
        self,
        walkers_starting_models: List[Any],
        perturbation_funcs: List[Callable[[Any], Tuple[Any, Number]]],
        log_likelihood_func: Callable[[Any], Number] = None,
        log_like_ratio_func: Callable[[Any, Any], Number] = None,
        n_chains: int = 10,
        n_cpus: int = 10,
        save_dpred: bool = True,
    ):
        self.walkers_starting_models = walkers_starting_models
        self.perturbation_funcs = [
            _preprocess_func(func) for func in perturbation_funcs
        ]
        if log_like_ratio_func is None:
            if log_likelihood_func is None:
                raise ValueError(
                    "at least one of `log_like_ratio_func` and `log_likelihood_func` needs"
                    "to be provided"
                )
            self.log_like_ratio_func = _LogLikeRatioFromFunc(log_likelihood_func)
        else:
            self.log_like_ratio_func = _preprocess_func(log_like_ratio_func)
        self.n_chains = n_chains
        self.n_cpus = n_cpus
        self.save_dpred = save_dpred
        self._chains = [
            BaseMarkovChain(
                id=i,
                starting_model=self.walkers_starting_models[i],
                perturbation_funcs=self.perturbation_funcs,
                log_like_ratio_func=self.log_like_ratio_func,
                save_dpred=self.save_dpred,
            )
            for i in range(n_chains)
        ]

    @property
    def chains(self) -> List[BaseMarkovChain]:
        """The ``MarkovChain`` instances of the current Bayesian inversion"""
        return self._chains

    def run(
        self,
        sampler: Sampler = None,
        n_iterations: int = 1000,
        burnin_iterations: int = 0,
        save_every: int = 100,
        verbose: bool = True,
        print_every: int = 100,
    ):
        r"""To run the inversion

        Parameters
        ----------
        sampler : bayesbay.samplers.Sampler, optional
            a sampler instance describing how chains intereact or modify their
            properties during sampling. This could be a sampler from the
            module :mod:`bayesbay.samplers` such as
            :class:`bayesbay.samplers.VanillaSampler` (default),
            :class:`bayesbay.samplers.ParallelTempering`, or
            :class:`bayesbay.samplers.SimulatedAnnealing`,
            or a customised sampler instance.
        n_iterations : int, optional
            total number of iterations to run, by default 1000
        burnin_iterations : int, optional
            the iteration number from which we start to save samples, by default 0
        save_every : int, optional
            the frequency with which we save the samples. By default a model is
            saved every after 100 iterations after the burn-in phase
        verbose : bool, optional
            whether to print the progress during sampling or not, by default True
        print_every : int, optional
            the frequency with which we print the progress and information during the
            sampling, by default 100 iterations
        """
        if sampler is None:
            sampler = VanillaSampler()
        sampler.initialize(self.chains)
        self._chains = sampler.run(
            n_iterations=n_iterations,
            n_cpus=self.n_cpus,
            burnin_iterations=burnin_iterations,
            save_every=save_every,
            verbose=verbose,
            print_every=print_every,
        )

    def get_results(
        self,
        keys: Union[str, List[str]] = None,
        concatenate_chains: bool = True,
    ) -> Union[Dict[str, list], list]:
        """To get the saved models

        Parameters
        ----------
        keys : Union[str, List[str]]
            one or more keys to retrieve from the saved models. This will be ignored when
            models are not of type :class:`State` or dict
        concatenate_chains : bool, optional
            whether to aggregate samples from all the Markov chains or to keep them
            seperate, by default True

        Returns
        -------
        Union[Dict[str, list], list]
            a dictionary from name of the attribute stored to the values, or a list of
            saved models
        """
        if isinstance(keys, str):
            keys = [keys]
        if hasattr(self.chains[0].saved_models, "items"):
            results_model = defaultdict(list)
            for chain in self.chains:
                for key, saved_values in chain.saved_models.items():
                    if keys is None or key in keys:
                        if concatenate_chains and isinstance(saved_values, list):
                            results_model[key].extend(saved_values)
                        else:
                            results_model[key].append(saved_values)
        else:
            results_model = []
            for chain in self.chains:
                if concatenate_chains:
                    results_model.extend(chain.saved_models)
                else:
                    results_model.append(chain.saved_models)
        return results_model


class BayesianInversion(BaseBayesianInversion):
    """A high-level class for performing Bayesian inversion using Markov Chain Monte
    Carlo (McMC) methods.

    This is a subclass of :class:`BaseBayesianInversion`.

    This class provides the basic structure for setting up and running McMC sampling,
    given user-configured parameterization settings, data targets and corresponding
    forward functions.

    Parameters
    ----------
    parameterization : bayesbay.parameterization.Parameterization
        pre-configured parameterization. This includes information about the dimension,
        parameterization bounds and properties of unknown parameterizations
    targets : List[bayesbay.Target]
        a list of data targets
    fwd_functions : Callable[[bayesbay.State], np.ndarray]
        a lsit of forward functions corresponding to each data targets provided above.
        Each function takes in a model and returns a numpy array of data predictions.
    n_chains: int, 10 by default
        the number of chains in the McMC sampling
    n_cpus: int, 10 by default
        the number of CPUs available. This is usually set to be equal to the number of
        chains if there are enough CPUs
    """

    def __init__(
        self,
        parameterization: Parameterization,
        targets: List[Target],
        fwd_functions: List[Callable[[State], np.ndarray]],
        n_chains: int = 10,
        n_cpus: int = 10,
        save_dpred: bool = True,
    ):
        self.targets = targets if isinstance(targets, list) else [targets]
        if not isinstance(fwd_functions, list):
            fwd_functions = [fwd_functions]
        self.fwd_functions = [_preprocess_func(func) for func in fwd_functions]

        self.parameterization = parameterization
        self.n_chains = n_chains
        self.n_cpus = n_cpus
        self.save_dpred = save_dpred
        self._chains = [
            MarkovChain(
                id=i,
                parameterization=self.parameterization,
                targets=self.targets,
                fwd_functions=self.fwd_functions,
                saved_dpred=self.save_dpred,
            )
            for i in range(n_chains)
        ]



from typing import Union, List, Callable, Tuple, Any, Dict
from numbers import Number
from collections import defaultdict
import random
import math
import numpy
from ._log_likelihood import LogLikelihood
from ._target import Target
from .parameterization import Parameterization
from ._state import State
from .exceptions import DimensionalityException, ForwardException, UserFunctionException


class BaseMarkovChain:
    r"""
    Low-level interface for a Markov Chain.

    Instantiation of this class is usually done by :class:`BaseBayesianInversion`.

    Parameters
    ----------
    id : Union[int, str]
        an integer or a string representing the ID of the current chain. For display
        purposes only
    starting_model : Any
        starting model of the current chain
    perturbation_funcs : List[Callable[[Any], Tuple[Any, Number]]]
        a list of perturbation functions
    log_like_ratio_func: Callable[[Any, Any], Number], optional
        function that calculates that the log likelihood ratio
        :math:`\frac{p\left({{\bf d}_{obs} \mid  {\bf m'}}\right)}{p\left({{\bf d}_{obs} \mid  {\bf m}}\right)}`.
        It takes in two models (of consistent type as other arguments of this class)
        and returns a scalar corresponding to the log likelihood ratio. This is utilised in the
        inversion by default, and ``log_likelihood_func`` gets used instead only when
        this argument is None. Default to None
    temperature : int, optional
        used to temper the log likelihood, by default 1
    """

    def __init__(
        self,
        id: Union[int, str],
        starting_model: Any,
        perturbation_funcs: List[Callable[[Any], Tuple[Any, Number]]],
        log_like_ratio_func: Callable[[Any, Any], Number] = None,
        temperature: float = 1,
        save_dpred: bool = True,
    ):
        self.id = id
        self.current_model = starting_model
        self.perturbation_funcs = perturbation_funcs
        self.perturbation_types = [func.__name__ for func in perturbation_funcs]
        self._temperature = temperature
        self.log_like_ratio_func = log_like_ratio_func
        self.save_dpred = save_dpred
        self._init_statistics()
        self._init_saved_models()

    @property
    def temperature(self) -> Number:
        """The current temperature used in the log likelihood ratio"""
        return self._temperature

    @temperature.setter
    def temperature(self, value):
        self._temperature = value
        self.current_model.temperature = value

    @property
    def saved_models(self) -> Union[Dict[str, list], list]:
        """All the models saved so far"""
        return getattr(self, "_saved_models", None)

    @property
    def statistics(self):
        return self._statistics

    def _init_saved_models(self):
        if isinstance(self.current_model, (State, dict)):
            self._saved_models = defaultdict(list)
        else:
            self._saved_models = []

    def _init_statistics(self):
        self._statistics = {
            "current_misfit": float("inf"),
            "n_explored_models": defaultdict(int),
            "n_accepted_models": defaultdict(int),
            "n_explored_models_total": 0,
            "n_accepted_models_total": 0,
            "exceptions": defaultdict(int),
        }

    def _save_model(self):
        if isinstance(self.current_model, (State, dict)):
            for k, v in self.current_model.items():
                if (
                    "n_dimensions" not in k
                    or self.parameterization.parameter_spaces[
                        k[: -len(".n_dimensions")]
                    ].trans_d
                ):
                    self.saved_models[k].append(v)
            if self.save_dpred and "dpred" in self.current_model.cache:
                self.saved_models["dpred"].append(
                    self.current_model.load_cache("dpred")
                )
        else:
            self.saved_models.append(self.current_model)

    def _save_statistics(self, perturb_i, accepted):
        perturb_type = self.perturbation_types[perturb_i]
        self._statistics["n_explored_models"][perturb_type] += 1
        self._statistics["n_accepted_models"][perturb_type] += 1 if accepted else 0
        self._statistics["n_explored_models_total"] += 1
        self._statistics["n_accepted_models_total"] += 1 if accepted else 0

    def _print_statistics(self):
        head = f"Chain ID: {self.id}"
        head += f"\nTEMPERATURE: {self.temperature}"
        head += f"\nEXPLORED MODELS: {self.statistics['n_explored_models_total']}"
        _accepted_total = self.statistics["n_accepted_models_total"]
        _explored_total = self.statistics["n_explored_models_total"]
        acceptance_rate = _accepted_total / _explored_total * 100
        head += "\nACCEPTANCE RATE: %d/%d (%.2f %%)" % (
            _accepted_total,
            _explored_total,
            acceptance_rate,
        )
        print(head)
        print("PARTIAL ACCEPTANCE RATES:")
        _accepted_all = self.statistics["n_accepted_models"]
        _explored_all = self.statistics["n_explored_models"]
        for perturb_type in sorted(self.statistics["n_explored_models"]):
            _explored = _explored_all[perturb_type]
            _accepted = _accepted_all[perturb_type]
            acceptance_rate = _accepted / _explored * 100
            print(
                "\t%s: %d/%d (%.2f%%)"
                % (perturb_type, _accepted, _explored, acceptance_rate)
            )
        # print("NUMBER OF FWD FAILURES: %d" % self.statistics["n_fwd_failures_total"])

    def _log_likelihood_ratio(self, new_model):
        return self.log_like_ratio_func(self.current_model, new_model)

    def _next_iteration(self, save_model):
        _last_exception = None
        for i in range(500):
            # choose one perturbation function and type
            i_perturb = random.randint(0, len(self.perturbation_funcs) - 1)
            perturb_func = self.perturbation_funcs[i_perturb]

            # perturb and get the partial acceptance probability excluding log
            # likelihood ratio
            try:
                new_model, log_prob_ratio = perturb_func(self.current_model)
            except (DimensionalityException, UserFunctionException) as e:
                _last_exception = e
                self._statistics["exceptions"][e.__class__.__name__] += 1
                continue

            # calculate the log likelihood ratio
            try:
                log_likelihood_ratio = self._log_likelihood_ratio(new_model)
            except (ForwardException, UserFunctionException) as e:
                _last_exception = e
                self._statistics["exceptions"][e.__class__.__name__] += 1
                continue

            # decide whether to accept
            acceptance_probability = log_prob_ratio + log_likelihood_ratio
            accepted = acceptance_probability > math.log(random.random())
            if accepted:
                self.current_model = new_model

            # save statistics and current model
            self._save_statistics(i_perturb, accepted)
            if save_model and self.temperature == 1:
                self._save_model()
            return
        raise RuntimeError(
            f"Chain {self.id} failed in perturb or forward calculation for 500 times. "
            "See above for the last exception."
        ) from _last_exception

    def advance_chain(
        self,
        n_iterations: int = 1000,
        burnin_iterations: int = 0,
        save_every: int = 100,
        verbose: bool = True,
        print_every: int = 100,
        on_begin_iteration: Callable[["BaseMarkovChain"], None] = None,
        on_end_iteration: Callable[["BaseMarkovChain"], None] = None,
    ):
        """advance the chain for a given number of iterations

        Parameters
        ----------
        n_iterations : int, optional
            the number of iterations to advance, by default 1000
        burnin_iterations : int, optional
            the iteration number from which we start to save samples, by default 0
        save_every : int, optional
            the frequency in which we save the samples, by default 100
        verbose : bool, optional
            whether to print the progress during sampling or not, by default True
        print_every : int, optional
            the frequency with which we print the progress and information during the
            sampling, by default 100 iterations
        on_begin_iteration : Callable[["BaseMarkovChain"], None], optional
            customized function that's to be run at before an iteration
        on_end_iteration : Callable[["BaseMarkovChain"], None], optional
            customized function that's to be run at after an iteration

        Returns
        -------
        BaseMarkovChain
            the chain itself is returned
        """
        for i in range(1, n_iterations + 1):
            if i <= burnin_iterations:
                save_model = False
            else:
                save_model = not (i - burnin_iterations) % save_every

            on_begin_iteration(self)
            self._next_iteration(save_model)
            on_end_iteration(self)
            if verbose and not i % print_every:
                self._print_statistics()

        return self


class MarkovChain(BaseMarkovChain):
    """High-level interface for a Markov Chain.

    This is a subclass of :class:`BaseMarkovChain`.

    Instantiation of this class is usually done by :class:`BayesianInversion`.

    Parameters
    ----------
    id : Union[int, str]
        an integer or a string representing the ID of the current chain. For display
        purposes only
    parameterization : bayesbay.parameterization.Parameterization
        pre-configured parameterization. This includes information about the dimension,
        parameterization bounds and properties of unknown parameterizations
    targets : List[bayesbay.Target]
        a list of data targets
    fwd_functions : Callable[[bayesbay.State], np.ndarray]
        a lsit of forward functions corresponding to each data targets provided above.
        Each function takes in a model and produces a numpy array of data predictions.
    temperature : int, optional
        used to temper the log likelihood, by default 1
    """

    def __init__(
        self,
        id: Union[int, str],
        parameterization: Parameterization,
        targets: List[Target],
        fwd_functions: Callable[[State], numpy.ndarray],
        temperature: float = 1,
        saved_dpred: bool = True,
    ):
        self.id = id
        self.parameterization = parameterization
        self.log_like_ratio_func = LogLikelihood(
            targets=targets,
            fwd_functions=fwd_functions,
        )
        self._temperature = temperature
        self.save_dpred = saved_dpred
        self.initialize()
        self._init_perturbation_funcs()
        self._init_statistics()
        self._init_saved_models()

    def initialize(self):
        """Initialize the parameterization by defining a starting model."""
        self.current_model = self.parameterization.initialize()
        for target in self.log_like_ratio_func.targets:
            target.initialize(self.current_model)

    def _init_perturbation_funcs(self):
        funcs_from_parameterization = self.parameterization.perturbation_functions
        funcs_from_log_likelihood = self.log_like_ratio_func.perturbation_functions
        self.perturbation_funcs = (
            funcs_from_parameterization + funcs_from_log_likelihood
        )
        self.perturbation_types = [f.__name__ for f in self.perturbation_funcs]



from abc import abstractmethod, ABC
from typing import Tuple
from numbers import Number
from .._state import State


class Perturbation(ABC):
    """Base class for a perturbation callable
    
    Its subclasses represent different types of perturbations. The core functions of
    each subclass are :meth:`perturb` and :meth:`log_prior_ratio` and are to be called
    by :meth:`bayesbay.BaseMarkovChain.advance_chain` on each iteration.
    
    Users don't instantiate these perturbation classes directly; the instances are 
    generated by :class:`bayesbay.parameterization.Parameterization` and 
    :class:`bayesbay.LogLikelihood` instead.
    """
    @abstractmethod
    def perturb(self, state: State) -> Tuple[State, Number]:
        r"""proposes a new state from the given state and calculates the log of
        the corresponding partial acceptance probability 
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}
        
        where :math:`p({\bf d}_{obs})` denotes the observed data and 
        :math:`\mathbf{J}` the Jacobian of the transformation.                                                                                                                                                                                                                                                                                                                                   

        Parameters
        ----------
        state : State
            the given current state

        Returns
        -------
        Tuple[State, Number]
            the proposed new state and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        raise NotImplementedError

    def __call__(self, state: State) -> Tuple[State, Number]:
        """same as :meth:`perturb`

        Parameters
        ----------
        state : State
            the given current state

        Returns
        -------
        Tuple[State, Number]
            proposed new state and the partial acceptance probability excluding log 
            likelihood ratio for this perturbation
        """
        return self.perturb(state)

    @property
    def type(self) -> str:
        """type of current perturbation (i.e. class name)

        Returns
        -------
        str
            the name of this current class
        """
        return self.__class__.__name__

    @property
    def __name__(self) -> str:
        """same as :attr:`type`

        Returns
        -------
        str
            the name of this current class
        """
        return self.type



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec 21 15:56:00 2022

@author: fabrizio
"""

from abc import ABC, abstractmethod
from typing import Any, Callable, Tuple, Union, Dict
from numbers import Number
import random
from functools import partial
import math
import numpy as np
from .._utils_1d import interpolate_linear_1d


TWO_PI = 2 * math.pi
SQRT_TWO_PI = math.sqrt(TWO_PI)


class Parameter(ABC):
    """Base class for an unknown parameter"""

    def __init__(self, **kwargs):
        self._name = kwargs["name"]
        if "position" in kwargs and kwargs["position"] is not None:
            self.position = np.array(kwargs["position"], dtype=float)
        else:
            self.position = None
        assert "perturb_std" in kwargs
        self.init_params = kwargs
        
    @property
    def name(self) -> str:
        return self._name

    @abstractmethod
    def initialize(
        self, position: Union[np.ndarray, Number] = None
    ) -> Union[np.ndarray, Number]:
        r"""initializes the values of this (possibly position-dependent) 
        parameter

        Parameters
        ----------
        position : Union[np.ndarray, Number], optional
            the position (in the discretization domain) associated with the value, 
            None by default

        Returns
        -------
        Union[np.ndarray, Number]
            an array of values or one value corresponding to the given positions
        """
        raise NotImplementedError

    @abstractmethod
    def perturb_value(self, value: Number, position: Number = None) -> Tuple[Number, Number]:
        r"""perturbs the given value, in a way that may depend on the position 
        associated with such a parameter value, and calculates the log of the 
        corresponding partial acceptance probability 
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},

        Parameters
        ----------
        value : Number
            the current value to be perturbed from
        position : Number, optional
            the position (in the discretization domain) associated with the value, 
            None by default

        Returns
        -------
        Tuple[Number, Number]
            the perturbed value and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        raise NotImplementedError

    @abstractmethod
    def log_prior(self, value: Number, position: Number = None) -> Number:
        r"""calculates the log of the prior probability of occurrence of the
        given value, which may depend on the considered position

        Parameters
        ----------
        value : Number
            the value to calculate the probability density for
        position : Number, optional
            the position (in the discretization domain) associated with the value, 
            None by default
        
        Returns
        -------
        Number
            the log prior probability density
        """
        raise NotImplementedError

    def set_custom_initialize(
        self,
        initialize_func: Callable[
            ["Parameter", Union[np.ndarray, Number]], Union[np.ndarray, Number]
        ],
    ):
        r"""sets a custom initialization function

        Parameters
        ----------
        initialize_func: Callable[[bayesbay.parameters.Parameter, Union[np.ndarray, Number]], Union[np.ndarray, Number]]
            The function to use for initialization. This function should take no arguments.

        Examples
        --------
        .. code-block:: python

            def my_init(
                param: bb.parameters.Parameter,
                position: Union[np.ndarray, Number]
            ) -> Union[np.ndarray, Number]:
                print("This is my custom init!")

            my_param.set_custom_initialize(my_init)
        """
        if not callable(initialize_func):
            raise ValueError("initialize_func must be a callable function.")
        self.initialize = partial(self._initializer, initialize_func)

    def _initializer(self, initialize_func, *args, **kwargs):
        return initialize_func(self, *args, **kwargs)

    def get_perturb_std(
        self, position: Union[Number, np.ndarray] = None
    ) -> Union[Number, np.ndarray]:
        r"""get the standard deviation of the Gaussian used to perturb
        the parameter, which may be dependent on the position in the
        discretization domain
        
        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            the position in the discretization domain at which the standard 
            deviation of the Gaussian used to perturb the parameter is returned.
            Default is None

        Returns
        -------
        Union[Number, np.ndarray]
            standard deviation of the Gaussian used to perturb the parameter,
            possibly at the specified position(s)
        """
        if self.has_hyper_param("perturb_std"):
            return self.get_hyper_param("perturb_std", position)
        else:
            raise NotImplementedError("`get_perturb_std` needs to be implemented")

    def add_hyper_params(self, hyper_params: Dict[str, Union[Number, np.ndarray]]):
        r"""Sets the attributes from the given dict and checks for errors
        
        Parameters
        ----------
        hyper_params : Dict[str, Union[Number, np.ndarray]]
            dictionary of attributes to be set
        """
        msg_scalar_when_pos_none = "should be a scalar when `position` is `None`"
        msg_scalar_or_same_len = (
            "should either be a scalar or have the same length as `position`"
        )
        for name, param_val in hyper_params.items():
            if self.position is None:
                assert np.isscalar(param_val), f"`{name}` {msg_scalar_when_pos_none}"
            else:
                assert np.isscalar(param_val) or \
                    len(param_val) == len(self.position), \
                        f"`{name}` {msg_scalar_or_same_len}"
            if not np.isscalar(param_val):
                param_val = np.array(param_val, dtype=float)
            setattr(self, name, self._init_hyper_param(param_val))
    
    def has_hyper_param(self, hyper_param: str) -> bool:
        r"""Whether or not the :class:`Parameter` instance has the specified
        attribute
        
        Parameters
        ----------
        hyper_param : str
            the name of the attribute
        """
        return hasattr(self, hyper_param)

    def get_hyper_param(
            self, 
            hyper_param: str, 
            position: Union[Number, np.ndarray] = None
            ) -> Union[Number, np.ndarray]:
        r"""Retrieves the value corresponding to the specified attribute, which 
        may be a function of position
        
        Parameters
        ----------
        hyper_param : str
            the name of the attribute
        position : Union[np.ndarray, Number], optional
            the position (in the discretization domain) associated with the value, 
            None by default
            
        Returns
        -------
        Union[Number, np.ndarray]
            value corresponding to the specified attribute
        """
        _hyper_param = getattr(self, hyper_param)
        if np.isscalar(_hyper_param):
            return _hyper_param
        elif position is not None:
            return _hyper_param(position)
        else:
            raise ValueError(
                f"`{hyper_param}` is dependent on position but got None for it"
            )

    def _init_hyper_param(self, hyper_param: Union[Number, np.ndarray]):
        # to be called after self.position is assigned
        return (
            hyper_param
            if np.isscalar(hyper_param)
            else partial(interpolate_linear_1d, x=self.position, y=hyper_param)
        )

    def __repr__(self) -> str:
        string = "%s(" % self.init_params["name"]
        for k, v in self.init_params.items():
            if k == "name":
                continue
            string += "%s=%s, " % (k, v)
        string = string[:-2]
        return string + ")"


class UniformParameter(Parameter):
    r"""Class for defining a free parameter according to a uniform probability
    distribution

    Parameters
    ----------
    name : str
        name of the parameter, for display and storing purposes
    vmin : Union[Number, np.ndarray]
        the lower bound for this parameter. This can either be a scalar or an array
        if the defined probability distribution is a function of ``position``
        in the discretization domain
    vmax : Union[Number, np.ndarray]
        the upper bound for this parameter. This can either be a scalar or an array
        if the defined probability distribution is a function of ``position``
        in the discretization domain
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the parameter. 
        This can either be a scalar or an array if the defined probability 
        distribution is a function of ``position`` in the discretization domain
    position : np.ndarray, optional
        position in the discretization domain, used to define a position-dependent
        probability distribution. None by default
    """

    def __init__(
        self,
        name: str,
        vmin: Union[Number, np.ndarray],
        vmax: Union[Number, np.ndarray],
        perturb_std: Union[Number, np.ndarray],
        position: np.ndarray = None,
    ):
        super().__init__(
            name=name,
            position=position,
            vmin=vmin,
            vmax=vmax,
            perturb_std=perturb_std,
        )
        self.add_hyper_params({
            "vmin": vmin, 
            "vmax": vmax, 
            "perturb_std": perturb_std, 
            "delta": np.array(vmax) - np.array(vmin), 
        })

    def get_delta(
        self, position: Union[Number, np.ndarray] = None
    ) -> Union[Number, np.ndarray]:
        r"""get the range :math:`\Delta v = v_{max} - v_{min}`, which may be
        dependent on the specified position in the discretization domain

        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            the position in the discretization domain at which the uniform 
            distribution range will be returned. None by default

        Returns
        -------
        Union[Number, np.ndarray]
            the range :math:`\Delta v = v_{max} - v_{min}`
        """
        return self.get_hyper_param("delta", position)

    def get_vmin_vmax(
        self, position: Union[Number, np.ndarray] = None
    ) -> Union[Tuple[Number, Number], Tuple[np.ndarray, np.ndarray]]:
        r"""get the lower and upper bounds of the parameter, which may be 
        dependent on position in the discretization domain

        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            the position in the discretization domain at which the the lower 
            and upper bounds of the parameter will be returned. None by default

        Returns
        -------
        Tuple[float, float]
            the lower (``vmin``) and upper (``vmax``)
        """
        vmin = self.get_hyper_param("vmin", position)
        vmax = self.get_hyper_param("vmax", position)
        return vmin, vmax

    def initialize(
        self, position: Union[np.ndarray, Number] = None
    ) -> Union[np.ndarray, Number]:
        r"""initialize the parameter, possibly at a specific position in the 
        discretization domain
        
        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            the position in the discretization domain at which the parameter is 
            initialized. None by default

        Returns
        -------
        Union[Number, np.ndarray]
            random number(s) chosen uniformly in the range associated with the
            given position(s)
        """
        vmin, vmax = self.get_vmin_vmax(position)
        if isinstance(position, Number):
            return random.uniform(vmin, vmax)
        else:
            n_vals = len(position) if position is not None else 1
            return np.random.uniform(vmin, vmax, n_vals)

    def perturb_value(self, value: Number, position: Number = None) -> Tuple[Number, Number]:
        r"""perturbs the given value, in a way that may depend on the position 
        in the discretization domain and calculates the log of the corresponding 
        partial acceptance probability,
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},

        which in this case equals zero. In the above equation, :math:`\bf m'` denotes 
        the perturbed model as obtained through a random deviate from a normal 
        distribution :math:`\mathcal{N}(v, \sigma)`, where :math:`v` denotes 
        the original ``value`` and :math:`\sigma` the standard deviation of the 
        Gaussian used for the perturbation. :math:`\sigma` may be dependent
        on the specified position (:attr:`UniformParameter.perturb_std`).

        Parameters
        ----------
        value : Number
            the current value to be perturbed from
        position : Number, optional
            the position in the discretization domain at which the parameter 
            ``value`` is to be perturbed

        Returns
        -------
        Tuple[Number, Number]
            the perturbed value and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert) = 0`
        """
        # randomly perturb the value until within range
        std = self.get_perturb_std(position)
        vmin, vmax = self.get_vmin_vmax(position)
        while True:
            random_deviate = random.normalvariate(0, std)
            new_value = value + random_deviate
            if new_value >= vmin and new_value <= vmax:
                return new_value, 0

    def log_prior(self, 
                  value: Number, 
                  position: Union[Number, np.ndarray] = None) -> Number:
        r"""calculates the log of the prior probability density for the given 
        value, which may be dependent on the position in the discretization
        domain

        Parameters
        ----------
        value : Number
            the value to calculate the probability density for
        position : Number, optional
            the position in the discretization domain at which the prior
            probability of the parameter ``value`` is to be retrieved

        Returns
        -------
        Number
            the log of the prior probability :math:`p(v) = \frac{1}{\Delta v}`,
            where :math:`v` denotes the ``value`` passed by the user and 
            :math:`\Delta v = v_{max} - v_{min}` the range within which the
            uniform parameter is allowed to vary. This may be
            dependent on position in the discretization domain
        """
        vmin, vmax = self.get_vmin_vmax(position)
        if vmin <= value <= vmax:
            return -math.log(vmax - vmin)
        else:
            return -math.inf


class GaussianParameter(Parameter):
    """Class for defining a free parameter using a Gaussian probability density
    function :math:`\mathcal{N}(\mu, \sigma)`, where :math:`\mu` denotes 
    the mean and :math:`\sigma` the standard deviation of the Gaussian

    Parameters
    ----------
    name : str
        name of the parameter, for display and storing purposes
    mean : Union[Number, np.ndarray]
        mean of the Gaussian. This can either be a scalar or an array
        if the defined probability distribution is a function of ``position``
        in the discretization domain
    std : Union[Number, np.ndarray]
        standard deviation of the Gaussian. This can either be a scalar or an array
        if the defined probability distribution is a function of ``position``
        in the discretization domain
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the parameter. 
        This can either be a scalar or an array if the defined probability distribution 
        is a function of ``position`` in the discretization domain
    position : np.ndarray, optional
        position in the discretization domain, used to define a position-dependent
        probability distribution. None by default
    """

    def __init__(self, name, mean, std, perturb_std, position=None):
        super().__init__(
            name=name,
            position=position,
            mean=mean,
            std=std,
            perturb_std=perturb_std,
        )
        self.add_hyper_params({
            "mean": mean, 
            "std": std, 
            "perturb_std": perturb_std, 
        })
    
    def get_mean(
        self, position: Union[Number, np.ndarray] = None
    ) -> Union[Number, np.ndarray]:
        """get the mean of the Gaussian parameter, which may be dependent on
        position in the discretization domain

        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            position(s) in the discretization domain at which the mean of the 
            Gaussian will be returned. None by default

        Returns
        -------
        Union[Number, np.ndarray]
            mean at the given position(s)
        """
        return self.get_hyper_param("mean", position)

    def get_std(
        self, position: Union[Number, np.ndarray] = None
    ) -> Union[Number, np.ndarray]:
        """get the standard deviation of the Gaussian parameter, which may be 
        dependent on position in the discretization domain

        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            position(s) in the discretization domain at which the standard 
            deviation of the Gaussian will be returned. None by default

        Returns
        -------
        Union[Number, np.ndarray]
            standard deviation at the given position(s)
        """
        return self.get_hyper_param("std", position)

    def initialize(
        self, position: Union[np.ndarray, Number] = None
    ) -> Union[np.ndarray, Number]:
        r"""initialize the parameter, possibly at a specific position in the 
        discretization domain
        
        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            the position in the discretization domain at which the parameter is 
            initialized. None by default

        Returns
        -------
        Union[np.ndarray, Number]
            an array of values or one value corresponding to the given position(s),
            chosen according to the normal distribution defined
            by :attr:`mean` and :attr:`std` at the given position(s)
        """
        mean = self.get_mean(position)
        std = self.get_std(position)
        n_vals = len(position) if position is not None else 1
        values = np.random.normal(mean, std, n_vals)
        return values

    def perturb_value(self, 
                      value: Number, 
                      position: Union[Number, np.ndarray] = None) -> Tuple[Number, Number]:
        r"""perturbs the given value, in a way that may depend on the position 
        in the discretization, and calculates the log of the corresponding 
        partial acceptance probability,
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},

        where :math:`\bf m'` denotes the perturbed model as obtained through a 
        random deviate from a normal distribution :math:`\mathcal{N}(v, \theta)`, 
        where :math:`v` denotes the original ``value`` and :math:`\theta` the 
        standard deviation of the Gaussian used for the perturbation. 
        :math:`\theta` may be dependent on the specified position 
        (:attr:`GaussianParameter.perturb_std`).
        
        Parameters
        ----------
        value : Number
            the current value to be perturbed from
        position : Number
            the position of the value to be perturbed
        
        Returns
        -------
        Tuple[Number, Number]
            the perturbed value and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        perturb_std = self.get_perturb_std(position)
        random_deviate = random.normalvariate(0, perturb_std)
        new_value = value + random_deviate
        mean = self.get_mean(position)
        std = self.get_std(position)
        ratio = (value - new_value) * (value + new_value - 2 * mean) / (2 * std**2)
        return new_value, ratio
    
    def log_prior(self, 
                  value: Number, 
                  position: Union[Number, np.ndarray] = None) -> Number:
        r"""calculates the log of the prior probability density for the given 
        value, which may be dependent on the position in the discretization
        domain

        Parameters
        ----------
        value : Number
            the value to calculate the probability density for
        position : Number, optional
            the position in the discretization domain at which the prior
            probability of the parameter ``value`` is to be retrieved
        
        Returns
        -------
        Number
            the log of the prior probability :math:`p(v) = \frac{1}{\sigma \sqrt{2 \pi}} 
            \exp \Big \lbrace -\frac{\left( v - \mu \right)^2}
            {2\sigma^2} \Big \rbrace`, where :math:`\mu` and :math:`\sigma`
            denote the (prior) mean and standard deviation of the 
            Gaussian parameter and :math:`v` the ``value`` at which the prior
            probability is to be retrieved. :math:`\mu` and :math:`\sigma` may
            be dependent on position in the discretization domain.
        """
        mean = self.get_mean(position)
        std = self.get_std(position)
        return -0.5 * np.log(2 * np.pi) - np.log(std) - 0.5 * ((value - mean) / std)**2


class CustomParameter(Parameter):
    """Class enabling the definition of an arbitrary prior for a free parameter

    Parameters
    ----------
    name : str
        name of the current parameter, for display and storing purposes
    log_prior : Callable[[Number, Number], Number]
    initialize : Callable[[np.ndarray], np.ndarray]
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the parameter. 
        This can either be a scalar or an array if the defined probability distribution 
        is a function of ``position`` in the discretization domain
    position : np.ndarray, optional
        position in the discretization domain, used to define a position-dependent
        probability distribution. None by default
    """
    def __init__(
        self,
        name: str,
        log_prior: Callable[[Number, Number], Number],
        initialize: Callable[[np.ndarray], np.ndarray],
        perturb_std: Union[Number, np.ndarray], 
        position: np.ndarray = None, 
    ):
        super().__init__(
            name=name,
            log_prior=log_prior,
            initialize=initialize,
            perturb_std=perturb_std, 
            position=position, 
        )
        self.add_hyper_params({
            "perturb_std": perturb_std
        })
        self._log_prior = log_prior
        self._initialize = initialize
        
    def initialize(
        self, position: Union[np.ndarray, Number] = None
    ) -> Union[np.ndarray, Number]:
        r"""initialize the parameter, possibly at a specific position in the 
        discretization domain
        
        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            the position in the discretization domain at which the parameter is 
            initialized. None by default
        
        Returns
        -------
        Union[np.ndarray, Number]
            an array of values or one value corresponding to the given positions
        """
        return self._initialize(position)
    
    def perturb_value(self, 
                      value: Number, 
                      position: Union[Number, np.ndarray] = None
                      ) -> Tuple[Number, Number]:
        r"""perturbs the given value, in a way that may depend on the position 
        in the discretization domain, and calculates the log of the corresponding 
        partial acceptance probability,
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},

        where :math:`\bf m'` denotes the perturbed model as obtained through a 
        random deviate from a normal distribution :math:`\mathcal{N}(v, \theta)`, 
        where :math:`v` denotes the original ``value`` and :math:`\theta` the 
        standard deviation of the Gaussian used for the perturbation. 
        :math:`\theta` may be dependent on the specified position in the 
        discretization domain (:attr:`CustomParameter.perturb_std`).
        
        Parameters
        ----------
        value : Number
            the current value to be perturbed from
        position : Number, optional
            the position of the value to be perturbed

        Returns
        -------
        Tuple[Number, Number]
            the perturbed value and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        perturb_std = self.get_perturb_std(position)
        random_deviate = random.normalvariate(0, perturb_std)
        new_value = value + random_deviate
        ratio = self.log_prior(new_value, position) - self.log_prior(value, position)
        return new_value, ratio

    def log_prior(self, 
                  value: Number, 
                  position: Union[Number, np.ndarray] = None) -> Number:
        r"""calculates the log of the prior probability density for the given 
        value, which may be dependent on the position in the discretization
        domain

        Parameters
        ----------
        value : Number
            the value to calculate the probability density for
        position : Number, optional
            the position in the discretization domain at which the prior
            probability of the parameter ``value`` is to be retrieved
        
        Returns
        -------
        Number
            the log prior probability density
        """
        try:
            return self._log_prior(value, position)
        except:
            return self._log_prior(value)



from abc import abstractmethod, ABC
from typing import Callable, List, Tuple
from functools import partial
import multiprocessing
import math
import random
import numpy as np

from .._markov_chain import MarkovChain, BaseMarkovChain


class Sampler(ABC):
    """Low-level class for defining the sampling criterion of a Markov chain
    and/or modifying its attributes in-between iterations
    """
    def __init__(self):
        self._extra_on_initialize = []
        self._extra_on_begin_iteration = []
        self._extra_on_end_iteration = []
        self._extra_on_end_advance_chain = []

    def initialize(self, chains: List[BaseMarkovChain]):  # called by external
        """initializes the given Markov Chains by calling :meth:`on_initialize`
        
        Parameters
        ----------
        chains : List[BaseMarkovChain]
            List of Markov chains used to sample the posterior
        """
        self.on_initialize(chains)
        self._chains = chains
        for func in self._extra_on_initialize:
            func(self, chains)

    def begin_iteration(self, chain: BaseMarkovChain):  # called by external
        """calls :meth:`on_begin_iteration` before beginning the current Markov
        chain iteration
        
        Parameters
        ----------
        chain : bayesbay.BaseMarkovChain
            Markov chain used to sample the posterior
        """
        self.on_begin_iteration(chain)
        for func in self._extra_on_begin_iteration:
            func(self, chain)

    def end_iteration(self, chain: BaseMarkovChain):  # called by external
        """calls :meth:`on_end_iteration` before passing to the next Markov
        chain iteration.
        
        Parameters
        ----------
        chain : bayesbay.BaseMarkovChain
            Markov chain used to sample the posterior
        """
        self.on_iteration_end(chain)
        for func in self._extra_on_end_iteration:
            func(self, chain)

    def end_advance_chain(self):  # called by external
        """calls :meth:`on_end_advance_chain` before concluding the batch of
        Markov chain iterations taking place after calling :meth:`advance_chain`
        """
        self.on_end_advance_chain()
        for func in self._extra_on_end_advance_chain:
            func(self)

    def add_on_initialize(
        self,
        func: Callable[["Sampler", List[MarkovChain]], None],
    ):
        """adds a custom function that will be called at the end of the chains
        initialization (see :meth:`initialize`)
        
        Parameters
        ----------
        func : Callable
            Function that takes as arguments a Sampler and a list of MarkovChains
        """
        self._extra_on_initialize.append(func)

    def add_on_begin_iteration(
        self, func: Callable[["Sampler", BaseMarkovChain], None]
    ):
        """adds a custom function that will be called at the beginning of each 
        Markov chain iteration (see :meth:`on_begin_iteration`)
        """
        self._extra_on_begin_iteration.append(func)

    def add_on_end_iteration(self, func: Callable[["Sampler", BaseMarkovChain], None]):
        """adds a custom function that will be called at the end of each Markov 
        chain iteration (see :meth:`on_end_iteration`)
        """
        self._extra_on_end_iteration.append(func)

    def add_on_end_advance_chain(self, func: Callable[["Sampler"], None]):
        """adds a custom function that will be called at the end of the batch of
        Markov chain iterations taking place after calling :meth:`advance_chain`
        """
        self._extra_on_end_advance_chain.append(func)

    @abstractmethod
    def on_initialize(self, chains):  # customized by subclass
        """defines the behaviour of the sampler at the initialization of the 
        Markov chains
        
        Parameters
        ----------
        chains : List[BaseMarkovChain]
            List of Markov chains used to sample the posterior
        """
        raise NotImplementedError

    @abstractmethod
    def on_begin_iteration(self, chain: BaseMarkovChain):  # customized by subclass
        """defines the behaviour of the sampler at the beginning of each
        Markov chain iteration
        
        Parameters
        ----------
        chain : BaseMarkovChain
            Markov chain used to sample the posterior
        """
        raise NotImplementedError

    @abstractmethod
    def on_end_iteration(self, chain: BaseMarkovChain):  # customized by subclass
        """defines the behaviour of the sampler at the end of each
        Markov chain iteration
        
        Parameters
        ----------
        chain : BaseMarkovChain
            Markov chain used to sample the posterior
        """
        raise NotImplementedError

    @abstractmethod
    def on_end_advance_chain(self):  # customized by subclass
        """defines the behaviour of the sampler at the end of the batch of
        Markov chain iterations taking place after calling :meth:`advance_chain`
        
        Parameters
        ----------
        chain : BaseMarkovChain
            Markov chain used to sample the posterior
        """
        raise NotImplementedError

    @abstractmethod
    def run(self):  # customized by subclass; called by external
        """function that allows for running the Bayesian inference by calling 
        :meth:`advance_chain`. 
        
        .. important::
            
            To work properly, a custom Sampler has to call within this function
            the method :meth:`advance_chain`.
        """
        raise NotImplementedError

    @property
    def chains(self) -> List[BaseMarkovChain]:
        """the ``MarkovChain`` instances of the current Bayesian inference"""
        return self._chains

    def advance_chain(
        self,
        n_iterations,
        n_cpus=10,
        burnin_iterations=0,
        save_every=100,
        verbose=True,
        print_every=100,
    ) -> List[BaseMarkovChain]:
        """advances the Markov chains for a given number of iterations

        Parameters
        ----------
        n_iterations : int
            the number of iterations to advance
        n_cpus : int, optional
            the number of CPUs to use
        burnin_iterations : int, optional
            the iteration number from which we start to save samples, by default 0
        save_every : int, optional
            the frequency in which we save the samples, by default 100
        verbose : bool, optional
            whether to print the progress during sampling or not, by default True
        print_every : int, optional
            the frequency with which we print the progress and information during the
            sampling, by default 100 iterations

        Returns
        -------
        List[BaseMarkovChain]
            the Markov chains
        """
        func = partial(
            MarkovChain.advance_chain,
            n_iterations=n_iterations,
            burnin_iterations=burnin_iterations,
            save_every=save_every,
            verbose=verbose,
            print_every=print_every,
            on_begin_iteration=self.on_begin_iteration,
            on_end_iteration=self.on_end_iteration,
        )
        if n_cpus > 1:
            pool = multiprocessing.Pool(n_cpus)
            self._chains = pool.map(func, self.chains)
            pool.close()
            pool.join()
        else:
            self._chains = [func(chain) for chain in self.chains]
        self.on_end_advance_chain()
        return self.chains


class VanillaSampler(Sampler):
    """High-level class to be used to sample the posterior by means of 
    reversible-jump Markov chain Monte Carlo.
    """
    def __init__(self):
        super().__init__()

    def on_initialize(self, chains: List[BaseMarkovChain]):
        pass

    def on_begin_iteration(self, chain: BaseMarkovChain):
        pass

    def on_end_iteration(self, chain: BaseMarkovChain):
        pass

    def on_end_advance_chain(self):
        pass

    def run(
        self,
        n_iterations,
        n_cpus=10,
        burnin_iterations=0,
        save_every=100,
        verbose=True,
        print_every=100,
    ) -> List[BaseMarkovChain]:
        return self.advance_chain(
            n_iterations=n_iterations,
            n_cpus=n_cpus,
            burnin_iterations=burnin_iterations,
            save_every=save_every,
            verbose=verbose,
            print_every=print_every,
        )


class ParallelTempering(Sampler):
    r"""High-level class to be used to sample the posterior by means of 
    reversible-jump Markov chain Monte Carlo accelerated with parallel tempering.

    See references below for details on parallel tempering [1]_, [2]_.
    
    Parameters
    ----------
    temperature_max : Number
        the maximum temperature attributed to the chains
    chains_with_unit_temperature : float
        the fraction of chains having unit temperature, 0.4 by default (i.e. 40%
        of the chains)
    swap_every : int
        the frequency with which the chain temperatures are randomly chosen 
        and possibly swapped during the sampling, by default 500 iterations
    
    References
    ----------
    .. [1] Ray et al. 2013, Robust and accelerated Bayesian inversion of marine 
        controlled-source electromagnetic data using parallel tempering
    .. [2] Sambridge 2014, A parallel tempering algorithm for probabilistic 
        sampling and multimodal optimization
    """
    def __init__(
        self,
        temperature_max=5,
        chains_with_unit_temperature=0.4,
        swap_every=500,
    ):
        super().__init__()
        self._temperature_max = temperature_max
        self._chains_with_unit_tempeature = chains_with_unit_temperature
        self._swap_every = swap_every

    def on_initialize(self, chains: List[BaseMarkovChain]):
        n_chains = len(chains)
        temperatures = np.ones(
            max(2, int(n_chains * self._chains_with_unit_tempeature)) - 1
        )
        if n_chains - temperatures.size > 0:
            size = n_chains - temperatures.size
        temperatures = np.concatenate(
            (temperatures, np.geomspace(1, self._temperature_max, size))
        )
        for i, chain in enumerate(chains):
            chain.temperature = temperatures[i]

    def on_begin_iteration(self, chain: BaseMarkovChain):
        pass

    def on_end_iteration(self, chain: BaseMarkovChain):
        pass

    def on_end_advance_chain(self):
        for i in range(len(self.chains)):
            chain1, chain2 = np.random.choice(self.chains, 2, replace=False)
            T1, T2 = chain1.temperature, chain2.temperature
            log_like_ratio = chain1._log_likelihood_ratio(chain2.current_model)
            prob = (1 / T1 - 1 / T2) * log_like_ratio
            if prob > math.log(random.random()):
                chain1.temperature = T2
                chain2.temperature = T1

    def run(
        self,
        n_iterations,
        n_cpus=10,
        burnin_iterations=0,
        save_every=100,
        verbose=True,
        print_every=100,
    ) -> List[BaseMarkovChain]:
        while True:
            iteration = self.chains[0].statistics["n_explored_models_total"]
            n_it = min(self._swap_every, n_iterations - iteration)
            burnin_it = max(
                0, min(self._swap_every, burnin_iterations - iteration)
            )
            self.advance_chain(
                n_iterations=n_it,
                n_cpus=n_cpus,
                burnin_iterations=burnin_it,
                save_every=save_every,
                verbose=verbose,
                print_every=print_every,
            )
            if iteration >= n_iterations:
                break
        return self.chains


class SimulatedAnnealing(Sampler):
    r"""High-level class to be used to sample the posterior by means of 
    reversible-jump Markov chain Monte Carlo accelerated with simulated annealing.

    See references below for details on simulated annealing [1]_.
    
    .. note::
        In our implementation, the temperature of each Markov chain is decreased
        exponentially with iteration, from :attr:`temperature_start` to 1, 
        during the burn-in phase. Using this ``SimulatedAnnealing`` is therefore 
        incompatible with setting ``burnin_iterations`` to zero in :meth:`run`.
    
    Parameters
    ----------
    temperature_start : Number
        the starting temperature of the Markov chains
        
        
    References
    ----------
    .. [1] Kirkpatrick et al. 1983, Optimization by simulated annealing
    """
    def __init__(self, temperature_start=10):
        super().__init__()
        self.temperature_start = temperature_start

    def on_initialize(self, chains: List[BaseMarkovChain]):
        for chain in chains:
            chain.temperature = self.temperature_start

    def on_begin_iteration(self, chain: BaseMarkovChain):
        iteration = chain.statistics["n_explored_models_total"]
        if iteration < self.burnin_iterations:
            chain.temperature = self.temperature_start * math.exp(
                -self.cooling_rate * iteration
            )
        elif iteration == self.burnin_iterations:
            chain.temperature = 1

    def on_end_iteration(self, chain: BaseMarkovChain):
        pass

    def on_end_advance_chain(self):
        pass

    def run(
        self,
        n_iterations,
        n_cpus=10,
        burnin_iterations=0,
        save_every=100,
        verbose=True,
        print_every=100,
    ) -> List[BaseMarkovChain]:
        self.burnin_iterations = burnin_iterations
        if burnin_iterations != 0:
            self.cooling_rate = math.log(self.temperature_start) / burnin_iterations
        self.advance_chain(
            n_iterations=n_iterations,
            n_cpus=n_cpus,
            burnin_iterations=burnin_iterations,
            save_every=save_every,
            verbose=verbose,
            print_every=print_every,
        )
        return self.chains



from abc import abstractmethod
from typing import Union, List, Tuple
from numbers import Number
import numpy as np

from ..parameters._parameters import Parameter
from ..parameterization._parameter_space import ParameterSpace
from .._state import ParameterSpaceState


class Discretization(Parameter, ParameterSpace):
    r"""Low-level class to define a discretization

    Parameters
    ----------
    name : str
        name of the discretization, for display and storing purposes
    spatial_dimensions : int
        number of spatial dimensions of the discretization, e.g. 1D, 2D, or 3D.
    vmin, vmax : Union[Number, np.ndarray]
        minimum/maximum value bounding each dimension
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the discretization
        in each dimension. 
    n_dimensions : Number, optional
        number of dimensions. None (default) results in a trans-dimensional
        discretization, with the dimensionality of the parameter space allowed
        to vary in the range ``n_dimensions_min`` - ``n_dimensions_max``
    n_dimensions_min, n_dimensions_max : Number, optional
        minimum and maximum number of dimensions, by default 1 and 10. These
        parameters are ignored if ``n_dimensions`` is not None, i.e. if the
        discretization is not trans-dimensional
    n_dimensions_init_range : Number, optional
        percentage of the range ``n_dimensions_min`` - ``n_dimensions_max`` used to
        initialize the number of dimensions (0.3. by default). For example, if 
        ``n_dimensions_min`` = 1, ``n_dimensions_max`` = 10, and 
        ``n_dimensions_init_range`` = 0.5,
        the maximum number of dimensions at the initialization is
        
        .. code-block:: python
            
            int((n_dimensions_max - n_dimensions_min) * n_dimensions_init_range + n_dimensions_max)
            
    parameters : List[Parameter], optional
        a list of free parameters, by default None
    birth_from : {"prior", "neighbour"}, optional
        whether to initialize the newborn basis functions by randomly drawing from 
        the prior (default) or by perturbing the neighbor one.
    """
    def __init__(
        self,
        name: str,
        spatial_dimensions: Number,
        perturb_std: Union[Number, np.ndarray] = 0.1,
        n_dimensions: int = None, 
        n_dimensions_min: int = 1, 
        n_dimensions_max: int = 10, 
        n_dimensions_init_range: Number = 0.3,
        parameters: List[Parameter] = None, 
        birth_from: str = "prior",
        **kwargs
    ):
        Parameter.__init__(
            self, 
            name=name,
            perturb_std=perturb_std,
            spatial_dimensions=spatial_dimensions,        
        )
        ParameterSpace.__init__(
            self, 
            name=name,
            n_dimensions=n_dimensions,
            n_dimensions_min=n_dimensions_min,
            n_dimensions_max=n_dimensions_max,
            n_dimensions_init_range=n_dimensions_init_range,
            parameters=parameters
        )
        self.spatial_dimensions = spatial_dimensions
        self.perturb_std = perturb_std
        self.birth_from = birth_from
        self._additional_kwargs = kwargs
    
    @abstractmethod
    def initialize(self, *args) -> ParameterSpaceState:
        """initializes the values of this discretization including its paramter values

        Returns
        -------
        ParameterSpaceState
            an initial parameter space state
        """
        raise NotImplementedError

    def _init_perturbation_funcs(self):
        raise NotImplementedError
        
    @abstractmethod
    def birth(
        self, param_space_state: ParameterSpaceState
    ) -> Tuple[ParameterSpaceState, Number]:
        r"""adds a dimension to the current parameter space and returns the
        thus obtained new state along with the log of the corresponding partial 
        acceptance probability,
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}.
            
        Parameters
        ----------
        ParameterSpaceState
            initial parameter space state

        Returns
        -------
        ParameterSpaceState
            new parameter space state
        Number
            log of the partial acceptance probability, 
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        raise NotImplementedError
    
    @abstractmethod
    def death(
        self, param_space_state: ParameterSpaceState
    ) -> Tuple[ParameterSpaceState, Number]:
        r"""removes a dimension from the given parameter space and returns the
        thus obtained new state along with the log of the corresponding partial 
        acceptance probability,
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}.
                    
        Parameters
        ----------
        ParameterSpaceState
            initial parameter space state

        Returns
        -------
        ParameterSpaceState
            new parameter space state
        Number
            log of the partial acceptance probability, 
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        raise NotImplementedError

    @abstractmethod
    def perturb_value(
        self, param_space_state: ParameterSpaceState, idimension: int
    ) -> Tuple[ParameterSpaceState, Number]:
        r"""perturbs the parameter space inherent to the discretization and 
        calculates the log of the corresponding partial acceptance probability 
        
        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} = 
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}} 
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}  
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}.
            
        .. note::
            All free parameters linked to the discretization will be perturbed
            at the specified index.

        Parameters
        ----------
        ParameterSpaceState
            initial parameter space state
        idimension : int
            index of the parameter-space dimension to be perturbed

        Returns
        -------
        Tuple[ParameterSpaceState, Number]
            the perturbed value and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m} 
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        raise NotImplementedError
    
    @abstractmethod
    def log_prior(self, param_space_state: ParameterSpaceState) -> Number:
        raise NotImplementedError

    def get_perturb_std(self, *args) -> Number:
        """get the standard deviation of the Gaussian used to perturb the 
        discretization
        """
        return self.perturb_std

    def _repr_dict(self) -> dict:      # to be called by ParameterSpace.__repr__
        attr_to_show = ParameterSpace._repr_dict(self)
        attr_to_show["spatial_dimensions"] = self.spatial_dimensions
        attr_to_show["perturb_std"] = self.perturb_std
        if self.trans_d:
            attr_to_show["birth_from"] = self.birth_from
        attr_to_show.update(self._additional_kwargs)
        return attr_to_show
    
    def __repr__(self) -> str:
        return ParameterSpace.__repr__(self)



import math
from typing import Any, List, Callable, Tuple
from numbers import Number
import numpy as np

from .exceptions import ForwardException
from ._state import State
from ._target import Target
from .perturbations._data_noise import NoisePerturbation
from ._utils import _preprocess_func


class LogLikelihood:
    """Helper class to evaluate the log likelihood ratio

    Parameters
    ----------
    targets : bayesbay.Target
        a list of data targets
    fwd_functions : List[Callable[[bayesbay.State], np.ndarray]]
        a list of forward functions corresponding to each data targets provided above.
        Each function takes in a model and produces a numpy array of data predictions.
    """

    def __init__(
        self,
        targets: List[Target],
        fwd_functions: List[Callable[[State], np.ndarray]],
    ):
        if not isinstance(fwd_functions, list):
            fwd_functions = [fwd_functions]
        if not isinstance(targets, list):
            targets = [targets]
        assert len(fwd_functions) == len(targets)

        self.targets = targets
        self.fwd_functions = [_preprocess_func(func) for func in fwd_functions]
        self._check_duplicate_target_names()
        self._init_perturbation_funcs()

    @property
    def perturbation_functions(self) -> List[Callable[[State], Tuple[State, Number]]]:
        """A list of perturbation functions associated with the data noise of the
        provided targets.

        Perturbation functions are included in this list only when the data noise of
        the target(s) is explicitly set to be unknown(s).
        """
        return self._perturbation_funcs

    def log_likelihood_ratio(self, old_state: State, new_state: State) -> Number:
        r"""Returns the (possibly tempered) log of the likelihood ratio

        .. math::
            \left[
                \frac{p\left(\mathbf{d}_{obs} \mid \mathbf{m'}\right)}{p\left(\mathbf{d}_{obs} \mid \mathbf{m}\right)}
            \right]^{\frac{1}{T}}
            =
            \left[
            \frac{\lvert \mathbf{C}_e \rvert}{\lvert \mathbf{C}^{\prime}_e \rvert}
            \exp\left(- \frac{\Phi(\mathbf{m'}) - \Phi(\mathbf{m})}{2}\right)
            \right]^{\frac{1}{T}},


        where :math:`\mathbf{C}_e` denotes the data covariance matrix,
        :math:`\Phi(\mathbf{m})` the data misfit associated with the model
        :math:`\mathbf{m}`, :math:`T` the chain temperature, and the prime
        superscript indicates that the model has been perturbed.


        Parameters
        ----------
        old_state : bayesbay.State
            the state of the Bayesian inference prior to the model perturbation
        new_state : bayesbay.State
            the state of the Bayesian inference after the model perturbation

        Returns
        -------
        Number
            log likelihood ratio
        """
        old_misfit, old_log_det = self._get_misfit_and_det(old_state)
        new_misfit, new_log_det = self._get_misfit_and_det(new_state)
        log_like_ratio = (old_log_det - new_log_det) + (old_misfit - new_misfit) / 2
        temperature = getattr(new_state, "temperature")
        return log_like_ratio / temperature

    def __call__(self, old_misfit, temperature) -> Any:
        return self.log_likelihood_ratio(old_misfit, temperature)

    def _check_duplicate_target_names(self):
        all_target_names = [t.name for t in self.targets]
        if len(all_target_names) != len(set(all_target_names)):
            raise ValueError("duplicate target names found")

    def _init_perturbation_funcs(self):
        hier_targets = [t for t in self.targets if t.is_hierarchical]
        self._perturbation_funcs = (
            [NoisePerturbation(self.targets)] if hier_targets else []
        )

    def _get_misfit_and_det(self, state: State) -> Tuple[Number, Number]:
        misfit = 0
        log_det = 0
        for target, fwd_func in zip(self.targets, self.fwd_functions):
            if state.has_cache("dpred"):
                dpred = state.load_cache("dpred")
            else:
                try:
                    dpred = fwd_func(state)
                except Exception as e:
                    raise ForwardException(e)
                state.store_cache("dpred", dpred)
            residual = dpred - target.dobs
            misfit += residual @ target.inverse_covariance_times_vector(state, residual)
            if target.is_hierarchical:
                log_det += target.log_determinant_covariance(state)
        if not isinstance(misfit, Number):
            raise TypeError(
                f"misfit is expected to be of a Number type, but is {type(misfit)} "
                "instead. This might be due to predicted data not being 1D. Try using "
                "numpy.squeeze() function on predicted data before returning it from "
                "your forward function"
            )
        return misfit, log_det



import math
import random
import numpy as np
import bayesbay as bb


uniform_param = bb.parameters.UniformParameter("uniform_param", -1, 1, 0.1)
gaussian_param = bb.parameters.GaussianParameter("gaussian_param", 0, 1, 0.1)
custom_param = bb.parameters.CustomParameter(
    name="custom_param",
    log_prior=lambda v: 0 if 0 <= v <= 1 else float("-inf"), 
    initialize=(lambda p: \
        np.random.uniform(0,1,len(p)) \
            if (not np.isscalar(p) and p is not None) \
                else random.uniform(0,1)), 
    perturb_std=1, 
)

my_ps = bb.parameterization.ParameterSpace(
    name="test", 
    n_dimensions=1, 
    parameters=[uniform_param, gaussian_param, custom_param], 
)

my_test_pss = bb.ParameterSpaceState(1, {
    "uniform_param": np.array([0,]), 
    "gaussian_param": np.array([0,]), 
    "custom_param": np.array([0.5,]), 
})

result = my_ps.log_prior(my_test_pss)
expected = math.log(1/2) + math.log(1/(math.sqrt(2*math.pi)*math.exp(0))) + 0
assert result == expected



from dataclasses import dataclass, field
from collections import namedtuple
from typing import Dict, Any, Union
from numbers import Number
import numpy as np


_DataNoiseState = namedtuple("DataNoiseState", ["std", "correlation"])


class DataNoiseState(_DataNoiseState):
    """Data structure that stores the state of the data noise parameters during
    the inference.
    """

    def copy(self) -> "DataNoiseState":
        """Returns a deep-copy of the current DataNoiseState

        Returns
        -------
        DataNoiseState
            the clone of self
        """
        return self._replace()

    def todict(self, name: str) -> dict:
        """Returns a dictionary containing the noise properties

        Parameters
        ----------
        name : str
            identifier for the ``DataNoiseState`` instance

        Returns
        -------
        dict
            dictionary object with keys

            - :attr:`name`.std
            - :attr:`name`.correlation (only returned if set as a free parameter,
              see :class:`Target`)
        """
        res = {f"{name}.std": self.std}
        if self.correlation is not None:
            res[f"{name}.correlation"] = self.correlation
        return res


@dataclass
class ParameterSpaceState:
    """Data structure that stores the state of a parameter space.

    Parameters
    ----------
    n_dimensions : int
        number of dimensions characterizing the parameter space
    param_values : Dict[str, Union[ParameterSpaceState, DataNoiseState]]
        dictionary containing parameter values, e.g.
        ``{"voronoi": ParameterSpaceState(3, {"voronoi": np.array([1,2,3]), "vs":
        np.array([4,5,6])}), "rayleigh": DataNoiseState(std=0.01,
        correlation=None)}``

    Raises
    ------
    TypeError
        when ``param_values`` is not a dict
    """

    n_dimensions: int
    param_values: Dict[str, np.ndarray] = field(default_factory=dict)

    def __post_init__(self):
        if not isinstance(self.n_dimensions, int):
            raise TypeError("n_dimensions should be an int")
        if not isinstance(self.param_values, dict):
            raise TypeError("param_values should be a dict")
        for name, values in self.param_values.items():
            if len(values) != self.n_dimensions:
                raise ValueError(
                    f"parameter {name} should have the same length as `n_dimensions` "
                    f"({self.n_dimensions}) but have {len(values)} instead"
                )
            self.set_param_values(name, values)

    def __getitem__(self, name: str) -> np.ndarray:
        return self.get_param_values(name)

    def set_param_values(self, param_name: str, values: np.ndarray):
        """Changes the numerical value(s) of a parameter

        Parameters
        ----------
        param_name : str
            the parameter name (i.e. the key in the ``param_values``)
        values : np.ndarray
            the value(s) to be set for the given ``param_name``
        """
        if isinstance(param_name, str):
            if not isinstance(values, np.ndarray):
                raise TypeError("parameter values should be a numpy ndarray instance")
            self.param_values[param_name] = values
            setattr(self, param_name, values)
        else:
            raise ValueError("`param_name` should be a string")

    def get_param_values(self, param_name: str) -> np.ndarray:
        """Get the value(s) of a parameter

        Parameters
        ----------
        param_name : str
            the parameter name (i.e. the key in the ``param_values`` dict)

        Returns
        -------
        Union[np.ndarray, None]
            the value(s) of the given ``param_name``, if present in
            :attr:`param_values`
        """
        if isinstance(param_name, str):
            return self.param_values.get(param_name, None)
        else:
            raise ValueError("`param_name` should be a string")

    def copy(self) -> "ParameterSpaceState":
        """Returns a clone of self

        Returns
        -------
        ParameterSpaceState
        """
        new_param_values = dict()
        for name, param_vals in self.param_values.items():
            new_param_values[name] = param_vals.copy()
        return ParameterSpaceState(self.n_dimensions, new_param_values)

    def todict(self, name: str) -> dict:
        """Returns a dictionary containing the numerical values defining the
        parameter space

        Parameters
        ----------
        name : str
            identifier for the ``ParameterSpaceState`` instance

        Returns
        -------
        dict
            dictionary object with keys

            - :attr:`name`.std
            - :attr:`name`.correlation (only returned if set as a free parameter,
              see :class:`Target`)
        """
        _discretization = "discretization"
        res = {f"{name}.n_dimensions": self.n_dimensions}
        res.update({k: v for k, v in self.param_values.items() if k != _discretization})
        if _discretization in self.param_values:
            res[f"{name}.{_discretization}"] = self.param_values[_discretization]
        return res


@dataclass
class State:
    """Data structure that stores a model state, including all the necessary
    information to perform the forward operation

    Parameters
    ----------
    param_values : Dict[str, Union[ParameterSpaceState, DataNoiseState]]
        dictionary containing parameter values, e.g.
        ``{"voronoi": ParameterSpaceState(3, {"voronoi": np.array([1,2,3]), "vs":
        np.array([4,5,6])}), "rayleigh": DataNoiseState(std=0.01,
        correlation=None)}``
    temperature : float
        the temperature of the Markov chain associated with this state
    cache : Dict[str, Any], optional
        cache for storing intermediate results
    extra_storage: Dict[str, Any], optional
        extra storage that will be saved into results (e.g. when one calls
        :meth:`BayesianInversion.get_results`)

    Raises
    ------
    TypeError
        when ``param_values`` is not a dict
    """

    param_values: Dict[str, Union[ParameterSpaceState, DataNoiseState]] = field(
        default_factory=dict
    )
    temperature: float = 1
    cache: Dict[str, Any] = field(default_factory=dict)
    extra_storage: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        if not isinstance(self.temperature, Number):
            raise TypeError("temperature should be a number")
        if not isinstance(self.param_values, dict):
            raise TypeError("param_values should be a dict")
        for name, values in self.param_values.items():
            self.set_param_values(name, values)

    def __getitem__(self, name: str) -> Union[ParameterSpaceState, DataNoiseState]:
        return self.get_param_values(name)

    def set_param_values(
        self, param_name: str, values: Union[ParameterSpaceState, DataNoiseState]
    ):
        """Changes the value(s) of a parameter

        Parameters
        ----------
        param_name : str
            the parameter name (i.e. the key in the ``param_values``)
        values : Union[ParameterSpaceState, DataNoiseState]
            the value(s) to be set for the given ``param_name``
        """
        if isinstance(param_name, str):
            if not isinstance(values, (ParameterSpaceState, DataNoiseState)):
                raise TypeError(
                    "parameter values should either be a ParameterSpaceState or a "
                    "`DataNoiseState` instance"
                )
            self.param_values[param_name] = values
            setattr(self, param_name, values)
        else:
            raise ValueError("`param_name` should be a string")

    def get_param_values(
        self, param_name: str
    ) -> Union[ParameterSpaceState, DataNoiseState]:
        """Get the value(s) of a parameter

        Parameters
        ----------
        param_name : str
            the parameter name (i.e. the key in the ``param_values`` dict)

        Returns
        -------
        Union[ParameterSpaceState, DataNoiseState]
            the value(s) of the given ``param_name``
        """
        if isinstance(param_name, str):
            return self.param_values.get(param_name, None)
        else:
            raise ValueError("`param_name` should be a string")

    def has_cache(self, name: str) -> bool:
        """Indicates whether there is cache value stored for the given ``name``

        Parameters
        ----------
        name : str
            the cache name to look up

        Returns
        -------
        bool
            whether there is cache stored for the given ``name``
        """
        return name in self.cache

    def load_cache(self, name: str) -> Any:
        """Load the cached value for the given ``name``

        Parameters
        ----------
        name : str
            the cache name to look up

        Returns
        -------
        Any
            the cache stored for the given ``name``
        """
        return self.cache[name]

    def store_cache(self, name: str, value: Any):
        """Store the given value to cache

        Parameters
        ----------
        name : str
            the cache name to store
        value : Any
            the cache value to store
        """
        self.cache[name] = value

    def _vars(self):
        all_vars = dict()
        for k, v in self.param_values.items():
            all_vars.update(v.todict(k))
        all_vars.update(self.extra_storage)
        return all_vars

    def __iter__(self):
        return iter(self._vars())

    def items(self):
        """Key-value pairs of all the values in the current model, expanding all
        parameter values, excluding cache

        Returns
        -------
        dict_items
            the key-value dict pairs of all the attributes
        """
        return self._vars().items()

    def copy(self, keep_dpred: bool = False) -> "State":
        """Creates a clone of the current State itself

        The following will be (deep-)copied over:

        - :attr:`param_values`
        - :attr:`temperature`

        And the following won't be copied at all:

        - :attr:`cache`
        - :attr:`extra_storage`

        If ``keep_dpred`` is ``True``, then the ``dpred`` in ``cache`` will be referred
        to in the new instance. Note that this is not a deep copy, since we assume no
        changes will be performed on predicted data for a certain state.

        Parameters
        ----------
        keep_dpred : bool, optional
            whether to copy over the predicted data (stored in cache)

        Returns
        -------
        State
            the clone of self
        """
        _param_values = dict()
        for k, v in self.param_values.items():
            _param_values[k] = v.copy()
        _state_args = {"param_values": _param_values, "temperature": self.temperature}
        if keep_dpred:
            _state_args["cache"] = {k: v for k, v in self.cache.items() if k == "dpred"}
        return State(**_state_args)



from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize
import numpy
import os


# Compiler arguments
extra_compile_args = ["-O3", "-ffast-math", "-march=native"]


def readme():
    with open("README.md", "r") as f:
        return f.read()

def read_version():
    version_file = os.path.join("src", "bayesbay", "_version.py")
    with open(version_file, "r") as f:
        lines = f.read()
    for line in lines.split("\n"):
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


ext_modules = [
    Extension(
        name="bayesbay._utils_1d",
        sources=["src/bayesbay/_utils_1d.pyx"],
        extra_compile_args=extra_compile_args,
        language="c++",
    )
]

setup(
    name="bayesbay",
    version=read_version(),
    long_description=readme(),
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        "numpy>=1.22",
        "matplotlib>=3.0.0",
    ],
    ext_modules=cythonize(ext_modules, language_level="3"),
    include_dirs=[numpy.get_include()],
)



from ._state import DataNoiseState, ParameterSpaceState, State
from ._target import Target
from ._log_likelihood import LogLikelihood
from ._markov_chain import MarkovChain, BaseMarkovChain
from ._bayes_inversion import BayesianInversion, BaseBayesianInversion

from . import samplers, parameters, perturbations, parameterization, discretization

from ._version import __version__


__all__ = [
    "DataNoiseState", 
    "ParameterSpaceState", 
    "State",
    "Target",
    "LogLikelihood",
    "MarkovChain",
    "BaseMarkovChain",
    "BayesianInversion",
    "BaseBayesianInversion",
]



class DimensionalityException(Exception):
    """
    Exception raised when trying to add/remove a dimension in a
    parameter space that has already reached the maximum/minimum number of
    allowed dimensions (see :class:`bayesbay.parameterization.ParameterSpace`).
    """

    def __init__(self, move):
        message = (
            "Error occured when trying to %s a Voronoi site."
            " The %s number of Voronoi cells had already been reached."
        )
        words = ("add", "maximum") if move == "Birth" else ("remove", "minimum")
        self.message = message % words
        super().__init__(self.message)

    def __str__(self):
        return self.message


class ForwardException(Exception):
    """
    Exception raised when a user-provided forward function raises an error
    """
    def __init__(self, original_exc):
        self.message = "error occurred when running the forward function - " + (
            original_exc.message
            if hasattr(original_exc, "message")
            else f"{type(original_exc).__name__}: {str(original_exc)}"
        )
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InitException(Exception):
    """
    Exception raised when users try to access a certain field that hasn't been
    intialized yet
    """
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class UserFunctionException(Exception):
    """Exception raised when a user-provided function raises an exception"""

    def __init__(self, original_exc):
        self.message = "error occurred when running the user-provided function - " + (
            original_exc.message
            if hasattr(original_exc, "message")
            else f"{type(original_exc).__name__}: {str(original_exc)}"
        )
        super().__init__(self.message)

    def __str__(self):
        return self.message
